constant __author = "Arne Goedeke <el@laramies.com>";
constant __version = "0.1";

//! This module implements the vCard standard as specified in rfc 2426. It is also
//! used to by @[Public.Standards.iCalendar] because the format is almost the same
//! in both cases. 
//! This module tries to be strict, i.e. it follows the rfc closely and - for instance
//! - will not parse vCards that use "\n" as line delimiter.
//!
//! @example
//! 	object parser = Public.Standards.vCard.strict_parser();
//! 	object vcard = parser->parse(data);
//! 	werror("name: %s\n", vcard["N"]->value);
//!	vcard["N"]->value = "John Doe";
//!	string s = vcard->render();
//! @note
//! 	Some vCard implementations do not follow the rfc. Therefore you might encounter
//! 	cases where "\n" is used as a line delimiter. Its not easily possible to work
//! 	around those cases automatically without breaking support in other cases. Hence
//! 	this parser will not accept that as input. 
string quote(string s) {
    int quote = 0;
    foreach (s;int i;int c) switch (c) {
    case ':':
    case ';':
    case ',':
	quote = 1;
	break;
    case '"':
    case 0..0x1f:
    case 0x7f:
	error("Control characters are not allowed inside vCard parameters.\n");
    }

    if (quote) return "\"" + s + "\"";

    return s;
}

//! Represents one line in a vCard format.
//! @seealso
//! 	@[Public.Standards.vCard.LineParser]
class ContentLine {
    //! Name of the content-line (e.g. "begin", "end", "N").
    string name;

    //! Parameters of the content-line. Type @expr{int@} is used for parameters that do
    //! not have a value.
    mapping(string:array(string)|string|int) params = ([]);

    //! Value of the content-line.
    array(string)|string value;

    string _sprintf(int type) {
	if (type == 'O') return sprintf("ContentLine(%s, %s, %O)", name, value, params);
	return "";
    }

    //! Render the content line. Folds lines according to the specification in rfc 2425.
    string render() {
	String.Buffer buf = String.Buffer();
	buf->add(name);

	foreach (params;string key;string|array(string)|int val) {
	    buf->putchar(';');
	    buf->add(key);

	    if (stringp(val)) {
		buf->putchar('=');
		buf->add(quote(val));
	    } else if (arrayp(val)) {
		buf->putchar('=');
		buf->add(map(val, quote)*",");
	    }
	}

	buf->putchar(':');
	buf->add(value);
	string ret = buf->get();

	if (sizeof(ret) > 75) {
	    return (ret / 72.0) * "\r\n ";
	}
	
	return ret;
    }
}

#if 0
class TypedProperties {
    mapping(string:object|array(object)) properties = ([]); 
    mapping(string:program) type_definitions;

    mixed `[](mixed key) {
	key = lower_case(key);
	mixed val = properties[key];

	if (!val) return val;
	
	if (objectp(val)) return val->get_value();

	if (!arrayp(val)) error("Bad entry in properties mapping.\n");
	
	array(mixed) ret = allocate(sizeof(val);
	foreach (val;int i;mixed typed) {
	   ret[i] = typed->get_value(); 
	}
    }

    void smart_add(string key, mixed value) {
	mixed current = properties[key];

	if (!current) {
	    properties[key] = value;
	} else if (arrayp(current)) {
	    properties[key] += ({ value });
	} else {
	    properties[key] = ({ current, value });
	}
    }

    mixed `[]=(mixed key, mixed value) {
	if (!stringp(key)) error("no non-string keys allowed in and iCalendar object.\n");
	key = lower_case(key);

	program type = type_definitions[key];

	if (!type) error("property %s not supported.\n", key);

	if (objectp(value)) {
	    if (Program.inherits(object_program(value), .Types.Base) && Program.inherits(object_program(value), type)) {
		smart_add(key, value);	
	    }
	} else smart_add(key, type(value));

	return value;
    }
}
#endif

class LowerCaseProperties {
    mapping(string:.ContentLine|array(.ContentLine)) properties = ([]);

    String.Buffer|string render(String.Buffer|void buffer) {
	String.Buffer buf = (buffer) ? buffer : String.Buffer();

	//foreach (properties;;.ContentLine|array(.ContentLine) val) {
	foreach (properties;;mixed val) {
	    if (arrayp(val)) {
		foreach (val;;.ContentLine v) {
		    buf->add(v->render());
		    buf->add("\r\n");
		}
	    } else {
		buf->add(val->render());
		buf->add("\r\n");
	    }
	}

	if (buffer) return buf;
	return buf->get();
    }

    // lower_case access to properties
    mixed `[](mixed key) {
	if (key && stringp(key)) {
	    return properties[lower_case(key)];
	}

	return properties[key];
    }

    mixed `[]=(mixed key, mixed val) {

	if (key && stringp(key)) {
	    mixed current = properties[key];

	    key = lower_case(key);

	    if (!current) {
		properties[key] = val;
	    } else if (arrayp(current)) {
		properties[key] += ({ val });
	    } else {
		properties[key] = ({ current, val });
	    }

	    return val;
	}

	error("Cannot store non-string keys in properties.\n");
    }
}

//! Simple class representing a vCard object.
class Object {
    inherit .LowerCaseProperties;

    //! Name of the object. Usually "VCARD".
    string name;

    //! Render the vCard to a string/String.Buffer.
    String.Buffer|string render(String.Buffer|void buffer) {
	String.Buffer buf = (buffer) ? buffer : String.Buffer();

	buf->add("BEGIN:"); buf->add(name); buf->add("\r\n");

	::render(buf);
	
	buf->add("END:"); buf->add(name); buf->add("\r\n");

	if (buffer) return buf;
	return buf->get();
    }

    string _sprintf(int type) {
	if (type == 'O') {
	    return sprintf("vCard.Component(%s, %O)\n", name, properties);
	}
    }
}

//! Create a new strict parser object. Use this instead of manually creating
//! an @[Public.Standards.Parser] instances unless you know what you are doing.
object strict_parser() {
    return .Parser(.Object);
}
